! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!          Utilities on top of fd4_domain and fd4_iter         !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Simple utilities on top of $fd4_domain and $fd4_iter.
!! Not optimized for performance.
!!
!! @see fd4_domain_mod
!! @see fd4_iter_mod
!! @author Matthias Lieber
module fd4_util_mod

  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_block_mod
  use fd4_domain_mod
  use fd4_iter_mod
  use fd4_mpi_mod
  use fd4_part_sfc_mod
  use fd4_balance_mod
  use fd4_mpitype_mod
  use fd4_comm_mod
  use rbtree_fd4_block_mod
  use stack_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
  
  ! public methods
  public :: fd4_util_put_array
  public :: fd4_util_put_array_uv
  public :: fd4_util_get_array
  public :: fd4_util_get_value
  public :: fd4_util_check_facevar
  public :: fd4_util_propagate_facevar
  public :: fd4_util_allocate_all_blocks
  
contains  


  include 'framework/fd4_domain.incf'
  include 'util/stack.incf'
  
  
  
  
  !! Allocate all blocks of the domain and keep them balanced over all processes.
  !!
  !! This function is only useful if the domain is not in *adaptive block mode*, i.e.
  !! no variable in the variable table has a threshold value.
  !! 
  !! Method: Simply mark all blocks for allocation and let $fd4_balance_readjust do
  !! the work.
  !! Calls $fd4_balance_readjust with Hilbert SFC load balancing and default parameters.
  !!
  !! *TODO*:
  !!  - does it work with ParMETIS?
  !!
  !! @see fd4_balance_mod
  subroutine fd4_util_allocate_all_blocks(domain, err)
    type(fd4_domain), intent(inout)         :: domain       !! the domain
    integer(i_k),     intent(out)           :: err          !! error status, 0 if ok
    
    integer(i_k) :: i, j, k
    
    err = -1
    
    call stack_create(domain%markedblocks, PRODUCT(domain%blnum(1:3)))
    
    ! we simply mark all blocks
    do k=1,domain%blnum(3)
      do j=1,domain%blnum(2)
        do i=1,domain%blnum(1)
          call stack_push( domain%markedblocks, fd4_domain_ijk2if(i,j,k) )
        end do
      end do
    end do
    
    domain%blocks_are_marked = .true.
    
    ! and call the load balancing routine
    call fd4_balance_readjust(domain, err, opt_force=.true.)

    call stack_delete(domain%markedblocks)
    
  end subroutine fd4_util_allocate_all_blocks
  
  
  !! Enforce consistency of the given face variable.
  !!
  !! Copies face values at the boundary of blocks to the corresponding neighbor block
  !! to ensure consistency. After calling to this subroutine, $fd4_util_check_facevar should
  !! report no problems.
  !!
  !! Example with 3 blocks:
  !!>
  !!     __    __    __
  !!    |  |  |  |  |  |
  !!    |__|  |__|  |__|
  !!    1  2  3  4  5  6
  !!<
  !!  - uppervalue = .true. propagates values from face 2 to face 3, 4 to 5, and 6 to 1 (if periodic)
  !!  - uppervalue = .false. propagates values from face 3 to face 2, 5 to 4, and 1 to 6 (if periodic)
  !!
  !! $fd4_util_propagate_facevar accepts arrays for $idx and $st.
  !! But all variables must be of the same face type!
  !!
  subroutine fd4_util_propagate_facevar(domain, idx, st, uppervalue, err)
    type(fd4_domain), intent(inout), target :: domain       !! the domain
    integer(i_k),     intent(in)            :: idx(:)       !! field indexes in vartab, must be of same face type!
    integer(i_k),     intent(in)            :: st(:)        !! time step indexes
    logical,          intent(in)            :: uppervalue   !! propagate upper value to neighbor block's lower value (or reverse?)
    integer(i_k),     intent(out)           :: err          !! error status, 0 if ok
    
    integer(i_k) :: face, bbnd(3,2), nbnd(3,2), dir, npos(3), fext(3), ext(0:3)
    integer(i_k) :: stride(size(idx),1:2), i, j, k, nvar
    integer(i_k), pointer :: bnd(:,:)
    integer, pointer :: bt(:,:,:,:)  ! block types
    integer(i8k) :: index
    type(fd4_iter), target :: iter
    type(fd4_block), pointer :: b, n
    type(fd4_mpi), pointer :: mpi
    type(fd4_mpitype), allocatable :: recvtype(:), sendtype(:)
    type(fd4_r_k_link), allocatable :: sendminptr(:), recvminptr(:)
    integer(MPI_ADDRESS_KIND) :: disp
    integer :: request, dest, blocktype
    logical :: is_min
    
    err = -1
    mpi => domain%mpi

    FD4_L1_ASSERT(size(idx)>size(st),'to few steps given')
    nvar = size(idx)

    face = domain%vartab(idx(1))%facevar
    FD4_L1_ASSERT(face<1.or.face>3,'no face variable at idx(1)')
    
    ! check if we have any blocks at all
    if(domain%blocks%length == 0) then
      err = 0
      return
    end if
    
    dir = 1
    if(uppervalue) dir = 2

    !! - prepare types for all possible block sizes -
    
    do i=1,nvar
      if(FD4_VERBOSE_LEVEL>=1 .and. domain%vartab(idx(i))%facevar/=face) then
        FD4_L1_ERROR('face types not matching at idx('//i2s(i)//')')
      end if
      stride(i,1:2) = domain%varinfo%fields(idx(i),st(i),1:2)
    end do

    fext = 0
    fext(face) = 1

    bnd => domain%mmbsz  ! min and max block size in domain

    ! allocate the index array for each possible block size
    allocate( bt( bnd(1,1):bnd(1,2),bnd(2,1):bnd(2,2),bnd(3,1):bnd(3,2),1:2 ) )

    ext(0) = domain%varinfo%field_len(face)
    bbnd(1:3,1) = 1

    ! loop over all possible block sizes
    do k=bnd(3,1),bnd(3,2)
      ext(3) = k + fext(3)
      do j=bnd(2,1),bnd(2,2)
        ext(2) = j + fext(2)
        do i=bnd(1,1),bnd(1,2)
          ext(1) = i + fext(1)
          ! construct the types for a block of size ext(:)
          bbnd(1:3,2) = ext(1:3)
          bbnd(face,1:2) = 1         ! only consider the first cell plane in face direction
          call fd4_comm_build_array_type(ext, stride, bbnd, bt(i,j,k,1), err)
          FD4_L3_ASSERT(err/=0,'fd4_comm_build_array_type failed')
          bbnd(face,1:2) = ext(face) ! only consider the additional cell plane in face direction
          call fd4_comm_build_array_type(ext, stride, bbnd, bt(i,j,k,2), err)
          FD4_L3_ASSERT(err/=0,'fd4_comm_build_array_type failed')
        end do
      end do
    end do

    allocate(recvtype(0:mpi%nprocs-1))
    allocate(sendtype(0:mpi%nprocs-1))

    allocate(sendminptr(0:mpi%nprocs-1))
    allocate(recvminptr(0:mpi%nprocs-1))

    mpi%requests%pos = 0


    ! iterate over existing blocks
    call fd4_iter_init(domain, iter)
    do while(associated(iter%cur))
      b => iter%cur
      
      !! - local copy -
      
      ! *upper* neighbor block present locally?
      n => b%neigh(face,2)%l
      if( associated(n) ) then
        ! neighbor block is a regular block (no ghost)
        if(n%pos(1)>0) then
          bbnd(:,1)=1
          bbnd(:,2)=b%ext
          nbnd(:,1)=1
          nbnd(:,2)=n%ext
          bbnd(face,:)=bbnd(face,2)+1 ! set to additional face plane
          nbnd(face,:)=1 ! set to first face plane
          if(uppervalue) then
#ifdef FD4_UTIL_PROPAGATE_FACEVAR_DEBUG
            write (1000+mpi%rank,'(A,3(I4),A,3(I4))') 'copy:',b%pos(1:3),' -> ',n%pos(1:3)
#endif
            do i=1,nvar
                n%fields(idx(i),st(i))%l(:,nbnd(1,1):nbnd(1,2),nbnd(2,1):nbnd(2,2),nbnd(3,1):nbnd(3,2)) & 
              = b%fields(idx(i),st(i))%l(:,bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2))
            end do
          else
#ifdef FD4_UTIL_PROPAGATE_FACEVAR_DEBUG
            write (1000+mpi%rank,'(A,3(I4),A,3(I4))') 'copy:',n%pos(1:3),' -> ',b%pos(1:3)
#endif
            do i=1,nvar
                b%fields(idx(i),st(i))%l(:,bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) &
              = n%fields(idx(i),st(i))%l(:,nbnd(1,1):nbnd(1,2),nbnd(2,1):nbnd(2,2),nbnd(3,1):nbnd(3,2))
            end do
          end if
        end if
      end if
      
      !! - copy from/to other ranks -
      
      ! 1: npos = the block I have to send data to
      npos = b%pos + fext * fd4_incdec(dir)
      dest = domain%blkmap(npos(1),npos(2),npos(3),domain%blkmap_now)
      if( dest>=0 .and. dest/=mpi%rank ) then
        ! put type to send data to dest into sendtype
        call MPI_Get_address( b%sdata(face)%l(1,1,1,1), disp, err)
        index = INT( fd4_domain_ijk2if(b%pos(1),b%pos(2),b%pos(3)), i8k)   ! index for sorting
        blocktype = bt(b%ext(1),b%ext(2),b%ext(3), dir) ! type for the block
#ifdef FD4_UTIL_PROPAGATE_FACEVAR_DEBUG
        write (1000+mpi%rank,'(A,3(I4),A,3(I4),A,I4,I10)') 'send:',b%pos(1:3),' -> ',npos(1:3),': ', dest, index
#endif
        call fd4_mpitype_add( sendtype(dest), index, 1, disp, blocktype, is_min)
        if(is_min) sendminptr(dest)%l => b%sdata(face)%l(1,1,1,1)
      end if
      
      ! 2: npos = the block I have to recv data from
      npos = b%pos - fext * fd4_incdec(dir)
      dest = domain%blkmap(npos(1),npos(2),npos(3),domain%blkmap_now)
      if( dest>=0 .and. dest/=mpi%rank ) then
        ! put type to receive data from dest into recvtype
        call MPI_Get_address( b%sdata(face)%l(1,1,1,1), disp, err)
        ! apply periodic boundary condition on npos, required for fd4_domain_ijk2if
        if(domain%periodic(face)) npos(face) = MOD( npos(face) + domain%blnum(face) - 1, domain%blnum(face) ) + 1
        index = INT( fd4_domain_ijk2if(npos(1),npos(2),npos(3)), i8k)   ! index for sorting
        blocktype = bt(b%ext(1),b%ext(2),b%ext(3), 3-dir) ! type for the block
#ifdef FD4_UTIL_PROPAGATE_FACEVAR_DEBUG
        write (1000+mpi%rank,'(A,3(I4),A,3(I4),A,I4,I10)') 'recv:',b%pos(1:3),' <- ',npos(1:3),': ', dest, index
#endif
        call fd4_mpitype_add( recvtype(dest), index, 1, disp, blocktype, is_min)
        if(is_min) recvminptr(dest)%l => b%sdata(face)%l(1,1,1,1)
      end if
        
      call fd4_iter_next(iter)
    end do
    
    
    !! - create send types and start sending messages -

    do dest=0,mpi%nprocs-1        

      if(sendtype(dest)%count>0) then
        ! commit the datatype
        call fd4_mpitype_commit(sendtype(dest), err)
        FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit failed')
        FD4_L3_ASSERT(.not.associated(sendminptr(dest)%l),'sendminptr not associated')
        ! send the data
        call MPI_Isend(sendminptr(dest)%l, 1, sendtype(dest)%datatype, dest, sendtype(dest)%count, mpi%comm, request, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
        call stack_push(mpi%requests, request)
#ifndef FD4_WAIT_BEFORE_TYPE_FREE
        call fd4_mpitype_free(sendtype(dest),.true.,.false.)
#endif      
      end if
    end do
    
    
    !! - create receive types and start receiving messages -

    do dest=0,mpi%nprocs-1

      if(recvtype(dest)%count>0) then
        ! commit the datatype
        call fd4_mpitype_commit(recvtype(dest), err)
        FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit failed')
        FD4_L3_ASSERT(.not.associated(recvminptr(dest)%l),'recvminptr not associated')
        ! recv the data
        call MPI_Irecv(recvminptr(dest)%l, 1, recvtype(dest)%datatype, dest, recvtype(dest)%count, mpi%comm, request, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Irecv failed')
        call stack_push(mpi%requests, request)
#ifndef FD4_WAIT_BEFORE_TYPE_FREE
        call fd4_mpitype_free(recvtype(dest),.true.,.false.)
#endif
      end if
    end do

    do k=bnd(3,1),bnd(3,2)
      do j=bnd(2,1),bnd(2,2)
        do i=bnd(1,1),bnd(1,2)
          call MPI_Type_free(bt(i,j,k,1), err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
          call MPI_Type_free(bt(i,j,k,2), err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
        end do
      end do
    end do
    deallocate(bt)

    !! finally wait for the communication to complete
    
    if(mpi%requests%pos>0) then
      call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
    end if    

#ifdef FD4_WAIT_BEFORE_TYPE_FREE
    do dest=0,mpi%nprocs-1
      call fd4_mpitype_free(recvtype(dest),.true.,.false.)
      call fd4_mpitype_free(sendtype(dest),.true.,.false.)
    end do
#endif

    ! free memory & types
    deallocate(recvtype)
    deallocate(sendtype)

    deallocate(sendminptr)
    deallocate(recvminptr)

    err = 0

  end subroutine fd4_util_propagate_facevar
  
  
  
  !! Check the given face variable for consistency.
  !!
  !! Checks if face values at opposite block boundaries are be equal.
  !! Reports to stdout.
  !!
  !! *Damn stupid implementation*: All blocks are migrated to rank 0 which performs the checks, than
  !! the blocks are distributed again over all ranks. *Use only for small debug runs*!
  !!
  subroutine fd4_util_check_facevar(domain, idx, st, err)
    type(fd4_domain), intent(inout)     :: domain       !! the domain
    integer(i_k),     intent(in)        :: idx          !! field index in vartab
    integer(i_k),     intent(in)        :: st           !! time step index
    integer(i_k),     intent(out)       :: err
    
    integer(i_k) :: face, bbnd(3,2), nbnd(3,2)
    type(fd4_iter), target :: iter
    type(fd4_block), pointer :: b, n

    face = domain%vartab(idx)%facevar
    FD4_L1_ASSERT(face<1.or.face>3,'no face variable '//i2s(idx))
    
    ! move all blocks to rank 0
    call fd4_balance_readjust(domain, err, FD4_BALANCE_SERIAL)
    FD4_L1_ASSERT(err/=0,'fd4_balance_readjust failed')
    
    
    ! iterate over existing blocks
    call fd4_iter_init(domain, iter)
    do while(associated(iter%cur))
      b => iter%cur
      ! neighbor block present?
      n => b%neigh(face,2)%l
      if( associated(n) ) then
        ! is neighbor block a regular block (no ghost)
        if(n%pos(1)>0) then
          bbnd(:,1)=1
          bbnd(:,2)=b%ext
          nbnd(:,1)=1
          nbnd(:,2)=n%ext
          bbnd(face,:)=bbnd(face,2)+1 ! set to additional face plane
          nbnd(face,:)=1 ! set to first face plane
          if(  ANY(    b%fields(idx,st)%l(:,bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2) ) &
                    /= n%fields(idx,st)%l(:,nbnd(1,1):nbnd(1,2),nbnd(2,1):nbnd(2,2),nbnd(3,1):nbnd(3,2) )   ) ) then
             ! error occured!
             write(fd4_out,'(A,A,I2,A,I4,I4,I4,A)') FD4_LINEHD,' face',face,' of block at',b%pos,' is not consistent!'
          end if
        end if
      end if      
      call fd4_iter_next(iter)
    end do
    
    
    ! load balance
    call fd4_balance_readjust(domain, err, FD4_BALANCE_HILBERT)
    FD4_L1_ASSERT(err/=0,'fd4_balance_readjust failed')
    
  end subroutine fd4_util_check_facevar
  


  !! Get a single value from the domain.
  !!
  !! *For debug use only*! Not optimized.
  function fd4_util_get_value(domain, idx, st, pos, root) result(value)
    type(fd4_domain), intent(in)        :: domain       !! the domain
    integer(i_k),     intent(in)        :: idx          !! field index in vartab
    integer(i_k),     intent(in)        :: st           !! time step index
    integer(i_k),     intent(in)        :: pos(0:3)     !! bin and position in global spatial coordinates
    integer(i_k),     intent(in)        :: root         !! write $array in rank $root
    real(r_k)                           :: value
    
    type(fd4_iter)  :: iter
    real(r_k) :: lv
    integer(i_k) :: err
    
    lv = 0.0_r_k
    value = 0.0_r_k

    ! iterate over existing blocks
    call fd4_iter_init(domain, iter)
    do while(associated(iter%cur))
      if( all(iter%cur%bnd(1:3,1) <= pos(1:3)) .and. all(iter%cur%bnd(1:3,2) >= pos(1:3)) ) then
        lv = iter%cur%fields(idx,st)%l(pos(0),pos(1)-iter%cur%bnd(1,1)+1,pos(2)-iter%cur%bnd(2,1)+1,pos(3)-iter%cur%bnd(3,1)+1)
      end if      
      call fd4_iter_next(iter)
    end do
    
    call MPI_Reduce(lv, value, 1, MPI_R_K, MPI_SUM, root, domain%mpi%comm, err )


  end function fd4_util_get_value


  !! Get array of domain size from domain's data structures.
  !!
  !! Optionally not setting array=0 in non-allocated blocks gives huge speed-up (factor 5 measured, due to cache).
  !!
  !! *TODO*
  !!  - face vars: take the value from a block that has a value, *not just simply cutting upper face*!!!
  !!    - currently, the very last value in each dimension in not put to the array, i.e. no face variable support
  !!  - $fd4_util_get to all ranks (no root)
  !!  - select more than one variable of same facevar type (see $fd4_ghostcomm_create, stride)
  !!  - subset on domain (as the original)
  subroutine fd4_util_get_array(domain, idx, st, abnd, array, err, root, novnull)
    type(fd4_domain), intent(in),target :: domain       !! the domain
    integer(i_k),     intent(in)        :: idx          !! field index in vartab
    integer(i_k),     intent(in)        :: st           !! time step index
    integer(i_k),     intent(in)        :: abnd(0:3,2)  !! bounds of the array: (bins, x, y, z)
    !! array with the data
    real(r_k),        intent(inout)     :: array(abnd(0,1):abnd(0,2),abnd(1,1):abnd(1,2),abnd(2,1):abnd(2,2),abnd(3,1):abnd(3,2))
    integer(i_k),     intent(out)       :: err          !! error status: 0...ok
    integer(i_k),     intent(in)        :: root         !! write $array in rank $root
    logical,optional, intent(in)        :: novnull      !! will not fill array for non-allocated blocks if set to .true. (*faster*)
    
    integer(i_k)    :: bbnd(3,2), facevar
    type(fd4_iter), target :: iter

    type(fd4_mpitype) :: sendtype
    type(fd4_mpitype), pointer :: recvtype(:)
    integer(i_k), pointer :: bnd(:,:)
    integer, pointer :: bt(:,:,:)  ! block types
    integer(i_k) :: ext(0:3), i, j, k, stride(1,2), sbnd(3,2), fext(3)
    integer(i8k) :: index
    type(fd4_block), pointer :: b
    type(fd4_r_k_link) :: minptr
    integer :: request, dest, blocktype
    logical :: is_min
    integer(MPI_ADDRESS_KIND) :: disp
    type(fd4_mpi), pointer :: mpi
    logical :: fill_with_vnull
    
    mpi => domain%mpi
    err = -1
    
#if ( FD4_VERBOSE_LEVEL >= 3 )
    if( abnd(0,2)-abnd(0,1)+1 < domain%vartab(idx)%nbins ) then
      FD4_L1_ERROR('abnd(0) does not match nbins of variable')
    end if
    if( abnd(1,2)-abnd(1,1) < domain%dbnd(1,2)-domain%dbnd(1,1) ) then
      FD4_L1_ERROR('abnd(1) too small')
    end if
    if( abnd(2,2)-abnd(2,1) < domain%dbnd(2,2)-domain%dbnd(2,1) ) then
      FD4_L1_ERROR('abnd(2) too small')
    end if
    if( abnd(3,2)-abnd(3,1) < domain%dbnd(3,2)-domain%dbnd(3,1) ) then
      FD4_L1_ERROR('abnd(3) too small')
    end if
    if(root<0 .or. root>=mpi%nprocs) then
      FD4_L1_ERROR('root out of bounds: '//i2s(root))
    end if
#endif

    mpi%requests%pos = 0
        
    if(mpi%rank==root) then
    
      !! receive data from all processes, including myself
      
      allocate(recvtype(0:mpi%nprocs-1))
      stride(1,1) = 1
      stride(1,2) = domain%vartab(idx)%nbins
      ext(0:3) = abnd(0:3,2) - abnd(0:3,1) + 1
      
      if(present(novnull)) then
        fill_with_vnull = .not. novnull
      else
        fill_with_vnull = .true.
      end if
      
      ! loop over all blocks of the domain
      do k=1,domain%blnum(3)
        do j=1,domain%blnum(2)
          do i=1,domain%blnum(1)
            
            index = INT( fd4_domain_ijk2if(i,j,k), i8k)   ! index for sorting
            dest = domain%blkmap(i,j,k,domain%blkmap_now)

            ! bounds of this non-existing block
            bbnd(1,:) = (/ domain%d(1)%bl(i) , domain%d(1)%bl(i+1) - 1 /)
            bbnd(2,:) = (/ domain%d(2)%bl(j) , domain%d(2)%bl(j+1) - 1 /)
            bbnd(3,:) = (/ domain%d(3)%bl(k) , domain%d(3)%bl(k+1) - 1 /)

            if(dest>=0) then
              ! bounds of subset in array which respects to the current block (asuming array starts with (1,1,1))
              sbnd(1:3,1) = bbnd(1:3,1) - domain%dbnd(1:3,1) + 1
              sbnd(1:3,2) = bbnd(1:3,2) - domain%dbnd(1:3,1) + 1
              ! build the datatype for array section sbnd
              call fd4_comm_build_array_type(ext, stride, sbnd, blocktype, err)
              FD4_L3_ASSERT(err/=0,'fd4_comm_build_array_type failed')
              ! note: displacement=0, because all blocktypes are relative to the beginning of array
              call fd4_mpitype_add(recvtype(dest), index, 1, 0_MPI_ADDRESS_KIND, blocktype)
            elseif(fill_with_vnull) then
              ! set array's data to vnull here
              array(stride(1,1):stride(1,2),bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) = domain%vartab(idx)%vnull
            end if
            
          end do
        end do
      end do
      
      
      do dest=0,mpi%nprocs-1        
        
        if(recvtype(dest)%count>0) then
          ! commit the datatype
          call fd4_mpitype_commit(recvtype(dest), err)
          FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit failed')
          ! recv the data directly into the array
          call MPI_Irecv(array(abnd(0,1),abnd(1,1),abnd(2,1),abnd(3,1)), &
                         1, recvtype(dest)%datatype, dest, recvtype(dest)%count, mpi%comm, request, err )
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Irecv failed')
          call stack_push(mpi%requests, request)
        end if
      
      end do      
    

      
    end if ! root recv
    
    

    !! send local blocks' data to root
    !! (even root does so - sends message to itself)
    
    ! check if we have any blocks at all
    if(domain%blocks%length > 0) then

      stride(1,1:2) = domain%varinfo%fields(idx,st,1:2)
      facevar = domain%vartab(idx)%facevar
      fext = 0
      if(facevar>0) fext(facevar) = 1

      bnd => domain%mmbsz  ! min and max block size in domain

      ! allocate the index array for each possible block size
      allocate( bt( bnd(1,1):bnd(1,2),bnd(2,1):bnd(2,2),bnd(3,1):bnd(3,2) ) )

      ext(0) = domain%varinfo%field_len(facevar)
      sbnd(1:3,1) = 1

      ! loop over all possible block sizes
      do k=bnd(3,1),bnd(3,2)
        ext(3) = k + fext(3)
        sbnd(3,2) = k
        do j=bnd(2,1),bnd(2,2)
          ext(2) = j + fext(2)
          sbnd(2,2) = j
          do i=bnd(1,1),bnd(1,2)
            ext(1) = i + fext(1)
            sbnd(1,2) = i
            ! construct the type for a block of size ext(:), take only the inner cells, i.e. without face
            call fd4_comm_build_array_type(ext, stride, sbnd, bt(i,j,k), err)
            FD4_L3_ASSERT(err/=0,'fd4_comm_build_array_type failed')
          end do
        end do
      end do

      ! iterate over existing blocks and create a sendtype
      call fd4_iter_init(domain, iter)
      do while(associated(iter%cur))
        b => iter%cur
        call MPI_Get_address( b%sdata(facevar)%l(1,1,1,1), disp, err)
        index = INT( fd4_domain_c2if(b%pos), i8k)   ! index for sorting
        blocktype = bt(b%ext(1),b%ext(2),b%ext(3)) ! type for the block
        !write (*,'(A,3(I3),A,4(I3))') 'ijk:',b%ext(1),b%ext(2),b%ext(3),' ubound:',ubound(b%sdata(facevar)%l)
        call fd4_mpitype_add( sendtype, index, 1, disp, blocktype, is_min)
        if(is_min) minptr%l => b%sdata(facevar)%l(1,1,1,1)
        call fd4_iter_next(iter)
      end do


      call fd4_mpitype_commit(sendtype, err)
      FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit failed')

      call MPI_Isend(minptr%l, 1, sendtype%datatype, root, sendtype%count, mpi%comm, request, err )
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
      call stack_push(mpi%requests, request)


      ! free all basic types and memory
      do k=bnd(3,1),bnd(3,2)
        do j=bnd(2,1),bnd(2,2)
          do i=bnd(1,1),bnd(1,2)
            call MPI_Type_free(bt(i,j,k), err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
          end do
        end do
      end do
      deallocate(bt)


    end if


    !! wait for the communication to complete
    
    if(mpi%requests%pos>0) then
      call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
    end if    


    !! free types used for communication
    
    if(sendtype%count>0) call fd4_mpitype_free(sendtype,.true.,.false.)
    if(mpi%rank==root) then
      do dest=0,mpi%nprocs-1        
        if(recvtype(dest)%count>0) call fd4_mpitype_free(recvtype(dest),.true.,.true.)
      end do
      deallocate(recvtype)
    end if
    
    err = 0

  end subroutine fd4_util_get_array






  !! Put array of given size to domain's data structures.
  !!
  !! *DEPRECATED, DO NOT USE*
  !!
  !! This suboutine is thought for filling data structures with initial data
  !! and not optimized for performance.
  !! Allocates new blocks when necessary. Will never allocate a
  !! block for a variable with FD4_NOTHRES.
  !!
  !! Currently, this is for testing. Maybe this subroutine will
  !! disappear in future.
  !!
  !! A safer way to put data into the framework are the coupling routines.
  !!
  !! @see fd4_couple_mod
  subroutine fd4_util_put_array(domain, idx, st, abnd, array, err, sbnd, opt_ignore_vthres)
    type(fd4_domain), intent(inout) :: domain       !! the domain
    integer(i_k),     intent(in)    :: idx          !! field index in vartab
    integer(i_k),     intent(in)    :: st           !! time step index
    integer(i_k),     intent(in)    :: abnd(0:3,2)  !! bounds of $array: (bins, x, y, z)
    !! array with the data
    real(r_k),        intent(in)    :: array(abnd(0,1):abnd(0,2),abnd(1,1):abnd(1,2),abnd(2,1):abnd(2,2),abnd(3,1):abnd(3,2))
    integer(i_k),     intent(out)   :: err          !! error status: 0...ok
    integer(i_k),optional,intent(in):: sbnd(0:3,2)  !! optionally one can specify a subset of $array to put into the domain
    logical,optional, intent(in)    :: opt_ignore_vthres !! ignore if FD4_NOTHRES is set and put everthing to domain
    
    integer(i_k)    :: i, x, y, z, bbnd(3,2), intersec(3,2), bpos(3), facevar
    type(fd4_iter)  :: iter
    integer(i_k), parameter :: shift(3) = (/1,1,1/)
    integer(i_k), allocatable :: newblocks(:,:), myblocks(:,:)
    integer(i_k) :: inewblocks, imyblocks
    integer(i_k) :: rbnd(0:3,2)
    logical :: ignore_vthres
    
    err = -1
    
    if(present(sbnd)) then
      rbnd=sbnd
    else
      rbnd=abnd
    end if
    
#if ( FD4_VERBOSE_LEVEL >= 3 )
    if( rbnd(0,2)-rbnd(0,1)+1 < domain%vartab(idx)%nbins ) then
      FD4_L1_ERROR('rbnd(0) does not match nbins of variable')
    end if
#endif
    
    if(present(opt_ignore_vthres) .and. domain%vartab(idx)%vthres == FD4_NOTHRES) then
      ignore_vthres = opt_ignore_vthres
    else
      ignore_vthres = .false.
    end if
    

    facevar = domain%vartab(idx)%facevar
    
    allocate(newblocks(3,PRODUCT(domain%blnum)))
    allocate( myblocks(3,PRODUCT(domain%blnum)))
    inewblocks=0

    ! no block creation for FD4_NOTHRES variables
    if(domain%vartab(idx)%vthres /= FD4_NOTHRES .or. ignore_vthres) then

      
      ! --- block lookup loop ---      
    
      ! loop over not existing blocks of the domain
      !TODO: limit loop bounds by abnd somehow
      do z=1,domain%blnum(3)
        do y=1,domain%blnum(2)
          do x=1,domain%blnum(1)
          
            bpos = (/ x, y, z /)

            if(domain%blkmap( x, y, z, domain%blkmap_now ) /= -1 ) then
              ! block exists (local or remote), consider it in newblock array
              inewblocks = inewblocks + 1 
              newblocks(:,inewblocks) = bpos(:)
              cycle
            end if 

            ! get bounds of the block
            do i=1,3
              bbnd(i,:) = (/ domain%d(i)%bl(bpos(i)) , domain%d(i)%bl(bpos(i)+1) - 1 /)
            end do

            ! calculate intersection of block with array
            intersec(1:3,1)=MAX(bbnd(1:3,1),rbnd(1:3,1))
            intersec(1:3,2)=MIN(bbnd(1:3,2),rbnd(1:3,2))
            
            ! check if intersection is not null
            if(intersec(1,1)<=intersec(1,2) .and. &
               intersec(2,1)<=intersec(2,2) .and. &
               intersec(3,1)<=intersec(3,2) ) then

              ! check if there's a value > threshold of this variable
              !TODO: can be optimized like in fd4_iter_finish
              !if( .true. .or. MAXVAL( array(:, intersec(1,1):intersec(1,2), & 
              if( ignore_vthres .or. &
                  MAXVAL( array(:, intersec(1,1):intersec(1,2), & 
                                   intersec(2,1):intersec(2,2), & 
                                   intersec(3,1):intersec(3,2)    ) )  > domain%vartab(idx)%vthres & 
                .or. &
                  MINVAL( array(:, intersec(1,1):intersec(1,2), & 
                                   intersec(2,1):intersec(2,2), & 
                                   intersec(3,1):intersec(3,2)    ) )  < -domain%vartab(idx)%vthres ) then

                ! add block to array
                inewblocks = inewblocks + 1 
                newblocks(:,inewblocks) = bpos(:)
              end if

            end if

          end do
        end do
      end do
      

      ! --- create new partition ---

      ! - takes already existing blocks into account
      ! - only adds new blocks into partitioning, does not touch already existing ones
      ! - the real balancing if performed at the end in fd4_balance_readjust (including migration, 
      !   which is left out here)     
      ! this routine returns the local partition in myblocks
      call fd4_part_new(domain, newblocks, inewblocks, myblocks, imyblocks, .true.)
      
      !print *,domain%mpi%rank,':',myblocks(:,1:imyblocks)

      !print *,domain%mpi%rank,domain%blkmap( 1,3,2, domain%blkmap_last ),domain%blkmap( 1,3,2, domain%blkmap_now )
      
      ! --- block creation loop ---

      do i=1,imyblocks
        ! check if the block exists anywhere else (in last partitioning)
        if( domain%blkmap( myblocks(1,i), myblocks(2,i), myblocks(3,i), domain%blkmap_last ) == -1 ) then
          ! create the block
          call fd4_domain_add_block(domain, myblocks(:,i), .true., err)
          ! err== 0 : block added -> ok
          ! err==-1 : block is already present -> also ok, because
          !           new block PLUS all existing blocks are in newblocks array
          FD4_L1_ASSERT(err>0,'cannot add block')
        end if
      end do


    end if


    ! --- data copy loop ---
    
    ! iterate over existing blocks and put the overlap to the block
    call fd4_iter_init(domain, iter)
    do while(associated(iter%cur))

      ! calculate intersection of block with array
      intersec(1:3,1)=MAX(iter%cur%bnd(1:3,1),rbnd(1:3,1))
      intersec(1:3,2)=MIN(iter%cur%bnd(1:3,2),rbnd(1:3,2))

      ! check if intersection is not null
      if(intersec(1,1)<=intersec(1,2) .and. &
         intersec(2,1)<=intersec(2,2) .and. &
         intersec(3,1)<=intersec(3,2) ) then
        
        ! if facevar: if not on upper face boundary: additional cell plane
        if( facevar /= 0 ) then
          if( rbnd(facevar,2) > iter%cur%bnd(facevar,2) ) then
            intersec(facevar,2) = intersec(facevar,2) + 1            
          end if
        end if

        ! transform the intersection to block local coordinates
        bbnd(:,1) = intersec(:,1) - iter%cur%bnd(:,1) + shift(:)
        bbnd(:,2) = intersec(:,2) - iter%cur%bnd(:,1) + shift(:)
        
        ! copy overlap to block's field
        iter%cur%fields(idx,st)%l(1:domain%vartab(idx)%nbins,  &
                                  bbnd(1,1):bbnd(1,2),           &
                                  bbnd(2,1):bbnd(2,2),           &
                                  bbnd(3,1):bbnd(3,2)    )  =  array(1:domain%vartab(idx)%nbins, &
                                                                     intersec(1,1):intersec(1,2),  & 
                                                                     intersec(2,1):intersec(2,2),  & 
                                                                     intersec(3,1):intersec(3,2)    )

        ! check if block is empty 
        if(domain%vartab(idx)%vthres/=FD4_NOTHRES) then
          call fd4_block_empty(iter%cur, domain%vartab, domain%varinfo, idx, st, domain%nghosts)
        end if

      end if
    
      call fd4_iter_next(iter)
    end do
    
    err = 0
    
    ! fd4_balance_readjust rebalances the load
    if(domain%vartab(idx)%vthres/=FD4_NOTHRES .or. ignore_vthres) then
    
      call fd4_balance_readjust(domain, err, opt_force=.true.)
      
    end if
    
  end subroutine fd4_util_put_array  


  !! Put u and v wind component to domain's data structures.
  !!
  !! *DEPRECATED, DO NOT USE*
  !!
  !! The arrays must be in full domain size.
  !! Array 1 must be face variable in 1st dimension (u).
  !! Array 2 must be face variable in 2nd dimension (v).
  !! This is a very specialized version of $fd4_util_put_array.
  !!
  subroutine fd4_util_put_array_uv(domain, idx1, st1, abnd1, array1, idx2, st2, abnd2, array2, err)
    type(fd4_domain), intent(inout) :: domain       !! the domain
    integer(i_k),     intent(in)    :: idx1          !! field index in vartab
    integer(i_k),     intent(in)    :: st1           !! time step index
    integer(i_k),     intent(in)    :: abnd1(0:3,2)  !! bounds of the array: (bins, x, y, z)
    !! array with the data
    real(r_k),        intent(in) :: array1(abnd1(0,1):abnd1(0,2),abnd1(1,1):abnd1(1,2),abnd1(2,1):abnd1(2,2),abnd1(3,1):abnd1(3,2))
    integer(i_k),     intent(in)    :: idx2          !! field index in vartab
    integer(i_k),     intent(in)    :: st2           !! time step index
    integer(i_k),     intent(in)    :: abnd2(0:3,2)  !! bounds of the array: (bins, x, y, z)
    !! array with the data
    real(r_k),        intent(in) :: array2(abnd2(0,1):abnd2(0,2),abnd2(1,1):abnd2(1,2),abnd2(2,1):abnd2(2,2),abnd2(3,1):abnd2(3,2))
    integer(i_k),     intent(out)   :: err          !! error status: 0...ok
    
    integer(i_k)    :: bbnd(3,2), intersec(3,2)
    type(fd4_iter)  :: iter

    err = -1
    
#if ( FD4_VERBOSE_LEVEL >= 3 )
    if( abnd1(0,2)-abnd1(0,1)+1 < domain%vartab(idx1)%nbins ) then
      FD4_L1_ERROR('abnd1(0) does not match nbins of variable')
    end if
    if( abnd2(0,2)-abnd2(0,1)+1 < domain%vartab(idx2)%nbins ) then
      FD4_L1_ERROR('abnd1(0) does not match nbins of variable')
    end if
    if(domain%vartab(idx1)%facevar/=1) then
      FD4_L1_ERROR('array1 not u')
    endif
    if(domain%vartab(idx2)%facevar/=2) then
      FD4_L1_ERROR('array2 not v')
    endif
    !TODO: check if abnd is full domain size ...
#endif


    ! --- data copy loop ---
    
    ! iterate over existing blocks and put the overlap to the block
    call fd4_iter_init(domain, iter)
    do while(associated(iter%cur))

      ! calculate intersection of block with array
      intersec(1:3,1)=iter%cur%bnd(1:3,1)
      intersec(1:3,2)=iter%cur%bnd(1:3,2)
        
        ! transform the intersection to block local coordinates
        !bbnd(:,1) = intersec(:,1) - iter%cur%bnd(:,1) + 1
        bbnd(:,2) = intersec(:,2) - iter%cur%bnd(:,1) + 1
        
        ! copy overlap to block's field
        iter%cur%fields(idx1,st1)%l(1:domain%vartab(idx1)%nbins,  &
                                  1:bbnd(1,2)+1,           &
                                  1:bbnd(2,2),           &
                                  1:bbnd(3,2)    )  =  array1(1:domain%vartab(idx1)%nbins, &
                                                                     intersec(1,1):intersec(1,2)+1,  & 
                                                                     intersec(2,1):intersec(2,2),  & 
                                                                     intersec(3,1):intersec(3,2)    )
        iter%cur%fields(idx2,st2)%l(1:domain%vartab(idx2)%nbins,  &
                                  1:bbnd(1,2),           &
                                  1:bbnd(2,2)+1,           &
                                  1:bbnd(3,2)    )  =  array2(1:domain%vartab(idx2)%nbins, &
                                                                     intersec(1,1):intersec(1,2),  & 
                                                                     intersec(2,1):intersec(2,2)+1,  & 
                                                                     intersec(3,1):intersec(3,2)    )

    
      call fd4_iter_next(iter)
    end do
    
  end subroutine fd4_util_put_array_uv




end module fd4_util_mod
