! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                       Merge sorted arrays                    !
!                                                              !
!--------------------------------------------------------------!

!! Merge multiple sorted integer arrays to one sorted array and remove duplicates.
!!
!! @author Matthias Lieber
module merge_mod

  use kinds_mod

  implicit none
  
contains

  !! Merge multiple sorted integer arrays to one sorted array and remove duplicates.
  !!
  !! - length of $displs = number of subarrays
  !! - length of $outarray should be large enough to hold the output (>=inarray is adequate)
  !!
  !! *TODO*:
  !! - performance
  !! - documentation
  subroutine merge_i(inarray, displs, outarray, outlength)
    integer(i_k), intent(in)    :: inarray(:)   !! an array containing sorted subarrays
    integer(i_k), intent(in)    :: displs(:)    !! array of displacements
    integer(i_k), intent(inout) :: outarray(:)  !! output array for merged result
    integer(i_k), intent(out)   :: outlength    !! number of elements written to $outarray
    
    integer(i_k) :: na, length, i, minv, amin, maxv
    integer(i_k) :: ptr(size(displs))
    
    na = size(displs)
    length = size(inarray)
    
    ! single subarray case
    if(na==1) then
      outlength=length
      outarray(1:length) = inarray(1:length)
      return
    end if
    
    ! check if the last subarray is 0 (its displacement is out of bounds)
    do while(displs(na) > length)
      na = na - 1
      if(na==1) then
        outlength=length
        outarray(1:length) = inarray(1:length)
        return
      end if
    end do

    ! reduce na to the last subarray containing values
    do while(displs(na-1)==displs(size(displs)))
      na = na - 1
      if(na==1) then
        outlength=length
        outarray(1:length) = inarray(1:length)
        return
      end if
    end do
    
    ! find the maxv
    ! (only search the last elements of the subarrays)
    maxv = inarray(length)
    do i=2,na
      if(displs(i)>1) maxv = MAX( maxv, inarray(displs(i)-1) )
    end do
    
    outlength = 0
    ptr = displs    !  pointer to current position in the subarrays
 
    ! find empty subarrays and set their ptr to 0
    do i=1,na-1
      if( displs(i+1) <= displs(i) ) ptr(i)=0
    end do
    
    ! check if the last subarray is 0 (its displacement is out of bounds)
    !if(ptr(na) > length) print *,'ptr(na) > length'
    !if(ptr(na) > length) ptr(na) = 0
    
    !print *,inarray
    !print *,displs(1:na),'(1:',na,')'
    !print *,ptr(1:na),'(1:',na,')'
    
    do while(.true.) 
    
      amin = -1
      minv = maxv
      do i=1,na
        if( ptr(i) /= 0 ) then
          if( inarray(ptr(i)) <= minv ) then
            amin = i
            minv = inarray(ptr(i))
          end if
        end if
      end do
       
      !print *,outlength,'amin, minv: ',amin,minv
       
      ! we found no element <= maxv: exit condition
      if(amin == -1) exit

      ptr(amin) = ptr(amin) + 1
      ! check if we are at the end of an input array
      if( amin == na .and. ptr(amin) >  length )  ptr(amin) = 0
      if( amin < na ) then
        if( ptr(amin) >= displs(amin+1) ) ptr(amin) = 0
      end if
      !if(  ( amin == na .and. ptr(amin) >  length ) .or.           &
      !     ( amin < na  .and. ptr(amin) >= displs(amin+1) ) ) then 
      !  ptr(amin) = 0
      !end if
      !print *,outlength,'ptr(amin): ',ptr(amin)

      ! ensure we have no duplicate
      if(outlength > 0) then      
        if(outarray(outlength) < minv) then
          ! put it to the output array
          outlength = outlength + 1
          outarray(outlength) = minv
        end if
      else
        ! put first element to the output array
        outlength = 1
        outarray(1) = minv
      end if

      !print *,outlength,'outarray(outlength): ',outarray(outlength)
      
    end do
       
    !print *,outarray(1:outlength)


  end subroutine merge_i


end module merge_mod
